
object CPointTester {
  
  // The main entry point of the tester
  
  def main (argv:Array[String]):Unit = test_all()

  
  //  Our own definition of equality
  //  Still depends on the provide isX() methods from implementation 
  
  private val tolerance : Double = 0.00001
  
  private def almost (d1 : Double, d2 : Double):Boolean = 
    (d1-d2 < tolerance) && (d1-d2 > -tolerance)
  
  private def eqCPoints (p : CPoint, q : CPoint):Boolean = {
    (almost(p.xCoord(),q.xCoord()) && 
     almost(p.yCoord(),q.yCoord()) &&
     p.color()==q.color())
  }
  
  private def normalize (angle:Double):Double = 
    if (angle >= 2*math.Pi)
      normalize(angle-2*math.Pi)
    else if (angle < 0)
      normalize(angle+2*math.Pi)
    else
      angle

  private def eqAngles (a : Double, b : Double):Boolean = 
    almost(normalize(a),normalize(b))


  // Create random points

  private val rnd = new scala.util.Random

  private def randomDouble ():Double = 
    1000 * rnd.nextDouble()

  private def randomAngle ():Double = 
    2 * math.Pi * rnd.nextDouble()

  private def randomColor ():Color = 
    if (rnd.nextBoolean())
      Color.red()
    else if (rnd.nextBoolean())
      Color.yellow()
    else
      Color.blue()
    
  private def randomCPoint ():CPoint = 
    CPoint.cartesian(randomDouble()-500,randomDouble()-500,randomColor())
      
    
  // These are individual tests, testing each equation in the specification
  // The format of the test name is 'test_Operation_Creator'
  // Each test takes values needed to create an instance of the ADT 
  //   via the requested creator
  // Each test returns 0 if the test succeeds and 1 if the test fails
  
  private def test_xCoord_cartesian (x:Double, y:Double, c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.xCoord()
    val exp = x
    assertT(almost(got,exp),
	    "cartesian("+x+","+y+","+c+").xCoord()",
	    exp,got)
  }

  private def test_yCoord_cartesian (x:Double, y:Double, c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.yCoord()
    val exp = y
    assertT(almost(got,exp),
	    "cartesian("+x+","+y+","+c+").yCoord()",
	    exp,got)
  }

  private def test_distanceFromOrigin_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.distanceFromOrigin()
    val exp = math.sqrt(x*x+y*y)
    assertT(almost(got,exp),
	    "cartesian("+x+","+y+","+c+").distanceFromOrigin()",
	    exp,got)
  }

  private def test_angleWithXAxis_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.angleWithXAxis()
    val exp = math.atan2(y,x)
    assertT(eqAngles(got,exp),
	    "cartesian("+x+","+y+","+c+").angleWithXAxis()",
	    exp,got)
  }

  private def test_color_cartesian (x:Double, y:Double, c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.color()
    val exp = c
    assertT(got==exp,
	    "cartesian("+x+","+y+","+c+").color()",
	    exp,got)
  }

  private def test_updateColor_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val nc = randomColor()
    val got = p.updateColor(nc)
    val exp = CPoint.cartesian(x,y,nc)
    assertT(eqCPoints(got,exp),
	    "cartesian("+x+","+y+","+c+").updateColor("+nc+")",
	    exp,got)
  }

  private def test_distance_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val q = randomCPoint()
    val got = p.distance(q)
    val exp = math.sqrt(math.pow(x-q.xCoord(),2)+
			math.pow(y-q.yCoord(),2))
    assertT(almost(got,exp),
	    "cartesian("+x+","+y+","+c+").distance("+q+")",
	    exp,got)
  }


  private def test_move_cartesian (x:Double, y:Double,c:Color):Int = {
    val p:CPoint = CPoint.cartesian(x,y,c)
    val dx = randomDouble()-500
    val dy = randomDouble()-500
    val got:CPoint = p.move(dx,dy)
    val exp = CPoint.cartesian(x+dx,y+dy,c)
    assertT(eqCPoints(got,exp),
	    "cartesian("+x+","+y+","+c+").move("+dx+","+dy+")",
	    exp,got)
  }

  private def test_add_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val q = randomCPoint()
    val got = p.add(q)
    val exp = CPoint.cartesian(x+q.xCoord(),y+q.yCoord(),q.color())
    assertT(eqCPoints(got,exp),
	    "cartesian("+x+","+y+","+c+").add("+q+")",
	    exp,got)
  }

  private def test_rotate_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val t = randomAngle()
    val got = p.rotate(t)
    val exp = CPoint.cartesian(x*math.cos(t)-y*math.sin(t),
                               x*math.sin(t)+y*math.cos(t),
			       c)
    assertT(eqCPoints(got,exp),
	    "cartesian("+x+","+y+","+c+").rotate("+t+")",
	    exp,got)
  }

  private def test_isEqual_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val q = randomCPoint()
    val got = p.isEqual(q)
    val exp = eqCPoints(p,q)
    assertT(got==exp,
	    "cartesian("+x+","+y+","+c+").isEqual("+q+")",
	    exp,got)
  }

  private def test_isOrigin_cartesian (x:Double, y:Double,c:Color):Int = {
    val p = CPoint.cartesian(x,y,c)
    val got = p.isOrigin()
    val exp = (almost(x,0) && almost(y,0))
    assertT(got==exp,
	    "cartesian("+x+","+y+","+c+").isOrigin()",
	    exp,got)
  } 


  // These are the tests for the creators
  // For each creator, we call the tests for operations on that creator
  // The result is the number of tests that failed
  
  val numberRandomCPoints = 100

  private def test_cartesian ():Int = {
    var failed : Int = 0
    for (i <- 1 to numberRandomCPoints) {
      val x = randomDouble()-500
      val y = randomDouble()-500
      val c = randomColor()
      try {
	failed += test_xCoord_cartesian(x,y,c)
	failed += test_yCoord_cartesian(x,y,c)
	failed += test_distanceFromOrigin_cartesian(x,y,c)
	failed += test_angleWithXAxis_cartesian(x,y,c)
	failed += test_color_cartesian(x,y,c)
	failed += test_updateColor_cartesian(x,y,c)
	failed += test_distance_cartesian(x,y,c)
	failed += test_move_cartesian(x,y,c)
	failed += test_add_cartesian(x,y,c)
	failed += test_add_cartesian(x,y,c)
	failed += test_rotate_cartesian(x,y,c)
	failed += test_isEqual_cartesian(x,y,c)
	failed += test_isOrigin_cartesian(x,y,c)
      } catch {
	case ex: RuntimeException => {
          // If there was an exception anywhere in there, then we
          //   have a problem
          assertT(false, "Exception: "+ex.getMessage(),"","")
	  return failed+1;
        }
      }
    }
    return failed
  }

  
  def test_all ():Unit = {
    var failed = 0;
    
    failed += test_cartesian()
    
    println("\nNumber of tests failed: " + failed)
  }
    



    /*   Result is expected to be true for passing tests, and false for
     *   failing tests.  If a test fails, we print out the provided
     *   message so the user can see what might have gone wrong. 
     *   
     * We return 0 if the test succeeded, and 1 if it failed 
     *
     * Be sure to review anything that doesn't make sense. */

    private def assertT[A] (result:Boolean, msg:String, exp:A, got:A):Int = {
        if (!result) {
          println("\n**TEST FAILED** "+ msg)
	  println("   Got: "+got)
	  println("   Expected: "+exp)
	}
	else
          print(".")
	return (if (result) 0 else 1)
    }



}
